@@ -44,9 +44,9 @@
 
 #include "dataprovider.h"
 
-#include <context.h>
-#include <data.h>
-#include <key.h>
+#include <gpgme++/context.h>
+#include <gpgme++/data.h>
+#include <gpgme++/key.h>
 
 using namespace QGpgME;
 using namespace GpgME;
@@ -106,11 +106,15 @@ static const char *originToString(Key::Origin origin)
 }
 
 static QGpgMEImportJob::result_type import_qba(Context *ctx, const QByteArray &certData, const QString &importFilter,
-                                               Key::Origin keyOrigin, const QString &keyOriginUrl)
+                                               const QStringList &importOptions, Key::Origin keyOrigin,
+                                               const QString &keyOriginUrl)
 {
     if (!importFilter.isEmpty()) {
         ctx->setFlag("import-filter", importFilter.toStdString().c_str());
     }
+    if (!importOptions.empty()) {
+        ctx->setFlag("import-options", importOptions.join(QLatin1Char{','}).toStdString().c_str());
+    }
     if (keyOrigin != Key::OriginUnknown) {
         if (const auto origin = originToString(keyOrigin)) {
             std::string value{origin};
@@ -149,27 +153,20 @@ static QGpgMEImportJob::result_type import_qba(Context *ctx, const QByteArray &c
 
 Error QGpgMEImportJob::start(const QByteArray &certData)
 {
-    run(std::bind(&import_qba, std::placeholders::_1, certData, importFilter(), keyOrigin(), keyOriginUrl()));
+    run(std::bind(&import_qba, std::placeholders::_1, certData, importFilter(), importOptions(), keyOrigin(), keyOriginUrl()));
     return Error();
 }
 
 GpgME::ImportResult QGpgME::QGpgMEImportJob::exec(const QByteArray &keyData)
 {
-    const result_type r = import_qba(context(), keyData, importFilter(), keyOrigin(), keyOriginUrl());
-    resultHook(r);
-    return mResult;
+    const result_type r = import_qba(context(), keyData, importFilter(), importOptions(), keyOrigin(), keyOriginUrl());
+    return std::get<0>(r);
 }
 
 Error QGpgMEImportJob::startLater(const QByteArray &certData)
 {
-    setWorkerFunction(std::bind(&import_qba, std::placeholders::_1, certData, importFilter(), keyOrigin(), keyOriginUrl()));
+    setWorkerFunction(std::bind(&import_qba, std::placeholders::_1, certData, importFilter(), importOptions(), keyOrigin(), keyOriginUrl()));
     return {};
 }
 
-// PENDING(marc) implement showErrorDialog()
-
-void QGpgME::QGpgMEImportJob::resultHook(const result_type &tuple)
-{
-    mResult = std::get<0>(tuple);
-}
 #include "qgpgmeimportjob.moc"
