@@ -60,7 +60,6 @@
 #include <shlobj.h>
 
 #include "util.h"
-#include "ath.h"
 #include "sema.h"
 #include "debug.h"
 #include "sys-util.h"
@@ -83,16 +82,27 @@
 # define GNUPG_REGKEY_3  "Software\\GnuPG"
 #endif
 
+/* Installation type constants.  */
+#define INST_TYPE_GPG4WIN  1
+#define INST_TYPE_GPGDESK  2
+
+/* Relative name parts for different installation types.  */
+#define INST_TYPE_GPG4WIN_DIR "\\..\\..\\GnuPG\\bin"
+#define INST_TYPE_GPGDESK_DIR "\\..\\GnuPG\\bin"
+
+
+
+
 DEFINE_STATIC_LOCK (get_path_lock);
 
 /* The module handle of this DLL.  If we are linked statically,
-   dllmain does not exists and thus the value of my_hmodule will be
+   dllmain does not exist and thus the value of my_hmodule will be
    NULL.  The effect is that a GetModuleFileName always returns the
    file name of the DLL or executable which contains the gpgme code.  */
 static HMODULE my_hmodule;
 
 /* These variables store the malloced name of alternative default
-   binaries.  The are set only once by gpgme_set_global_flag.  */
+   binaries.  They are set only once by gpgme_set_global_flag.  */
 static char *default_gpg_name;
 static char *default_gpgconf_name;
 /* If this variable is not NULL the value is assumed to be the
@@ -168,6 +178,48 @@ wchar_to_utf8 (const wchar_t *string)
 }
 
 
+/* Return a malloced wide char string from a UTF-8 encoded input
+   string STRING.  Caller must free this value. On failure returns
+   NULL; caller may use GetLastError to get the actual error number.
+   Calling this function with STRING set to NULL is not defined. */
+static wchar_t *
+utf8_to_wchar (const char *string)
+{
+  int n;
+  wchar_t *result;
+
+
+  n = MultiByteToWideChar (CP_UTF8, 0, string, -1, NULL, 0);
+  if (n < 0)
+    return NULL;
+
+  result = (wchar_t *) malloc ((n+1) * sizeof *result);
+  if (!result)
+    return NULL;
+
+  n = MultiByteToWideChar (CP_UTF8, 0, string, -1, result, n);
+  if (n < 0)
+    {
+      free (result);
+      return NULL;
+    }
+  return result;
+}
+
+
+/* Same as utf8_to_wchar but calling it with NULL returns
+   NULL.  So a return value of NULL only indicates failure
+   if STRING is not set to NULL. */
+static wchar_t *
+utf8_to_wchar0 (const char *string)
+{
+  if (!string)
+    return NULL;
+
+  return utf8_to_wchar (string);
+}
+
+
 /* Replace all forward slashes by backslashes.  */
 static void
 replace_slashes (char *string)
@@ -220,19 +272,19 @@ _gpgme_allow_set_foreground_window (pid_t pid)
 
   if (!pid || pid == (pid_t)(-1))
     {
-      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", NULL,
 	      "no action for pid %d", (int)pid);
     }
   else if (func)
     {
       int rc = func (pid);
-      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", NULL,
 	      "called for pid %d; result=%d", (int)pid, rc);
 
     }
   else
     {
-      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", NULL,
 	      "function not available");
     }
 #endif /* HAVE_ALLOW_SET_FOREGROUND_WINDOW */
@@ -245,7 +297,7 @@ void
 _gpgme_w32_cancel_synchronous_io (HANDLE thread)
 {
   static int initialized;
-  static BOOL (WINAPI * func)(DWORD);
+  static BOOL (WINAPI * func)(HANDLE);
   void *handle;
 
   if (!initialized)
@@ -268,13 +320,14 @@ _gpgme_w32_cancel_synchronous_io (HANDLE thread)
     {
       if (!func (thread) && GetLastError() != ERROR_NOT_FOUND)
         {
-          TRACE (DEBUG_ENGINE, "gpgme:CancelSynchronousIo", 0,
-                  "called for thread %p: ec=%d", thread, GetLastError ());
+          TRACE (DEBUG_ENGINE, "gpgme:CancelSynchronousIo", NULL,
+                 "called for thread %p: ec=%u",
+                 thread, (unsigned int)GetLastError ());
         }
     }
   else
     {
-      TRACE (DEBUG_ENGINE, "gpgme:CancelSynchronousIo", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:CancelSynchronousIo", NULL,
 	      "function not available");
     }
 }
@@ -395,7 +448,7 @@ find_program_in_dir (const char *dir, const char *name)
   if (!result)
     return NULL;
 
-  if (access (result, F_OK))
+  if (_gpgme_access (result, F_OK))
     {
       free (result);
       return NULL;
@@ -408,7 +461,7 @@ find_program_in_dir (const char *dir, const char *name)
 static char *
 find_program_at_standard_place (const char *name)
 {
-  char path[MAX_PATH];
+  wchar_t path[MAX_PATH];
   char *result = NULL;
 
   /* See https://wiki.tcl-lang.org/page/Getting+Windows+%22special+folders%22+with+Ffidl for details on compatibility.
@@ -416,20 +469,24 @@ find_program_at_standard_place (const char *name)
      We First try the generic place and then fallback to the x86
      (i.e. 32 bit) place.  This will prefer a 64 bit of the program
      over a 32 bit version on 64 bit Windows if installed.  */
-  if (SHGetSpecialFolderPathA (NULL, path, CSIDL_PROGRAM_FILES, 0))
+  if (SHGetSpecialFolderPathW (NULL, path, CSIDL_PROGRAM_FILES, 0))
     {
-      result = _gpgme_strconcat (path, "\\", name, NULL);
-      if (result && access (result, F_OK))
+      char *utf8_path = wchar_to_utf8 (path);
+      result = _gpgme_strconcat (utf8_path, "\\", name, NULL);
+      free (utf8_path);
+      if (result && _gpgme_access (result, F_OK))
         {
           free (result);
           result = NULL;
         }
     }
   if (!result
-      && SHGetSpecialFolderPathA (NULL, path, CSIDL_PROGRAM_FILESX86, 0))
+      && SHGetSpecialFolderPathW (NULL, path, CSIDL_PROGRAM_FILESX86, 0))
     {
-      result = _gpgme_strconcat (path, "\\", name, NULL);
-      if (result && access (result, F_OK))
+      char *utf8_path = wchar_to_utf8 (path);
+      result = _gpgme_strconcat (utf8_path, "\\", name, NULL);
+      free (utf8_path);
+      if (result && _gpgme_access (result, F_OK))
         {
           free (result);
           result = NULL;
@@ -489,6 +546,27 @@ _gpgme_set_override_inst_dir (const char *dir)
 }
 
 
+/* Used by gpgme_set_global_flag to set the installation type.
+ * VALUE is a string interpreted as integer with this meaning:
+ *   0 = standard
+ *   1 = Gpg4win 4 style (INST_TYPE_GPG4WIN)
+ *   2 = GnuPG (VS-)Desktop style (INST_TYPE_GPGDESK)
+ * If VALUE is NULL, nothing is changed.  The return value is the
+ * previous value.
+ */
+int
+_gpgme_set_get_inst_type (const char *value)
+{
+  static int inst_type;
+  int previous_type;
+
+  previous_type = inst_type;
+  if (value)
+    inst_type = atoi (value);
+  return previous_type;
+}
+
+
 /* Return the full file name of the GPG binary.  This function is used
    iff gpgconf was not found and thus it can be assumed that gpg2 is
    not installed.  This function is only called by get_gpgconf_item
@@ -532,13 +610,40 @@ _gpgme_get_gpg_path (void)
 
   /* 4. Print a debug message if not found.  */
   if (!gpg)
-    _gpgme_debug (DEBUG_ENGINE, -1, NULL, NULL, NULL,
+    _gpgme_debug (NULL, DEBUG_ENGINE, -1, NULL, NULL, NULL,
                   "_gpgme_get_gpg_path: '%s' not found", name);
 
   return gpg;
 }
 
 
+/* Helper for _gpgme_get_gpgconf_path.  */
+static char *
+find_version_file (const char *inst_dir)
+{
+  char *fname;
+
+  fname = _gpgme_strconcat (inst_dir, "\\..\\", "VERSION.sig", NULL);
+  if (fname && !_gpgme_access (fname, F_OK))
+    {
+      fname[strlen(fname)-4] = 0;
+      if (!_gpgme_access (fname, F_OK))
+        return fname;
+    }
+  free (fname);
+  /* Check the case that a binary in gnupg/bin uses libgpgme.  */
+  fname = _gpgme_strconcat (inst_dir, "\\..\\..\\", "VERSION.sig", NULL);
+  if (fname && !_gpgme_access (fname, F_OK))
+    {
+      fname[strlen(fname)-4] = 0;
+      if (!_gpgme_access (fname, F_OK))
+        return fname;
+    }
+  free (fname);
+  return NULL;
+}
+
+
 /* This function is only called by get_gpgconf_item and may not be
    called concurrently.  */
 char *
@@ -546,12 +651,53 @@ _gpgme_get_gpgconf_path (void)
 {
   char *gpgconf = NULL;
   const char *inst_dir, *name;
+  int inst_type;
+  char *dir = NULL;
 
   name = default_gpgconf_name? get_basename(default_gpgconf_name):"gpgconf.exe";
 
-  /* 1. Try to find gpgconf.exe in the installation directory of gpgme.  */
   inst_dir = _gpgme_get_inst_dir ();
-  if (inst_dir)
+  inst_type = _gpgme_set_get_inst_type (NULL);
+
+  /* 0.0. If no installation type has been explicitly requested guess
+   * one by looking at files used by the installation type.  */
+  if (inst_dir && !inst_type)
+    {
+      gpgrt_stream_t fp;
+      char buffer[128];
+      int n;
+
+      free (dir);
+      dir = find_version_file (inst_dir);
+      if (dir && (fp = gpgrt_fopen (dir, "r")))
+        {
+          n = gpgrt_fread (buffer, 1, 128, fp);
+          if (n > 10)
+            {
+              buffer[n-1] = 0;
+              if (strstr (buffer, "GnuPG") && strstr (buffer, "Desktop"))
+                inst_type = INST_TYPE_GPGDESK;
+            }
+          gpgrt_fclose (fp);
+        }
+    }
+
+  /* 0.1. If an installation type was requested or guessed try to find
+   * gpgconf.exe depending on that installation type.  */
+  if (inst_dir
+      && (inst_type == INST_TYPE_GPG4WIN || inst_type == INST_TYPE_GPGDESK))
+    {
+      free (dir);
+      dir = _gpgme_strconcat
+        (inst_dir,
+         inst_type == INST_TYPE_GPG4WIN? INST_TYPE_GPG4WIN_DIR
+         /*                         */ : INST_TYPE_GPGDESK_DIR,
+         NULL);
+      gpgconf = find_program_in_dir (dir, name);
+    }
+
+  /* 1. Try to find gpgconf.exe in the installation directory of gpgme.  */
+  if (!gpgconf && inst_dir)
     {
       gpgconf = find_program_in_dir (inst_dir, name);
     }
@@ -567,8 +713,7 @@ _gpgme_get_gpgconf_path (void)
   /* 3. Try to find gpgconf.exe using the Windows registry. */
   if (!gpgconf)
     {
-      char *dir;
-
+      free (dir);
       dir = read_w32_registry_string (NULL,
                                       GNUPG_REGKEY_2,
                                       "Install Directory");
@@ -586,10 +731,7 @@ _gpgme_get_gpgconf_path (void)
             }
         }
       if (dir)
-        {
-          gpgconf = find_program_in_dir (dir, name);
-          free (dir);
-        }
+        gpgconf = find_program_in_dir (dir, name);
     }
 
   /* 4. Try to find gpgconf.exe from Gpg4win below CSIDL_PROGRAM_FILES.  */
@@ -598,19 +740,28 @@ _gpgme_get_gpgconf_path (void)
       gpgconf = find_program_at_standard_place ("GNU\\GnuPG\\gpgconf.exe");
     }
 
-  /* 5. Try to find gpgconf.exe relative to us.  */
+  /* 5. Try to find gpgconf.exe relative to us as Gpg4win installs it.  */
   if (!gpgconf && inst_dir)
     {
-      char *dir = _gpgme_strconcat (inst_dir, "\\..\\..\\GnuPG\\bin", NULL);
-      gpgconf = find_program_in_dir (dir, name);
       free (dir);
+      dir = _gpgme_strconcat (inst_dir, INST_TYPE_GPG4WIN_DIR, NULL);
+      gpgconf = find_program_in_dir (dir, name);
     }
 
-  /* 5. Print a debug message if not found.  */
+  /* 6. Try to find gpgconf.exe relative to us as GnuPG VSD installs it. */
+  if (!gpgconf && inst_dir)
+    {
+      free (dir);
+      dir = _gpgme_strconcat (inst_dir, INST_TYPE_GPGDESK_DIR, NULL);
+      gpgconf = find_program_in_dir (dir, name);
+    }
+
+  /* Print a debug message if not found.  */
   if (!gpgconf)
-    _gpgme_debug (DEBUG_ENGINE, -1, NULL, NULL, NULL,
+    _gpgme_debug (NULL, DEBUG_ENGINE, -1, NULL, NULL, NULL,
                   "_gpgme_get_gpgconf_path: '%s' not found",name);
 
+  free (dir);
   return gpgconf;
 }
 
@@ -706,7 +857,7 @@ my_mkstemp (char *tmpl)
     random_time_bits = (((uint64_t)ft.dwHighDateTime << 32)
                         | (uint64_t)ft.dwLowDateTime);
   }
-  value += random_time_bits ^ ath_self ();
+  value += random_time_bits ^ ((uintptr_t)GetCurrentThreadId ());
 
   for (count = 0; count < attempts; value += 7777, ++count)
     {
@@ -781,6 +932,74 @@ _gpgme_mkstemp (int *fd, char **name)
 }
 
 
+/* Like access but using windows _waccess */
+int
+_gpgme_access (const char *path, int mode)
+{
+  wchar_t *u16 = utf8_to_wchar0 (path);
+  int r = _waccess (u16, mode);
+
+  free(u16);
+  return r;
+}
+
+
+/* Like CreateProcessA but mapping the arguments to wchar API */
+int
+_gpgme_create_process_utf8 (const char *application_name_utf8,
+                            char *command_line_utf8,
+                            LPSECURITY_ATTRIBUTES lpProcessAttributes,
+                            LPSECURITY_ATTRIBUTES lpThreadAttributes,
+                            BOOL bInheritHandles,
+                            DWORD dwCreationFlags,
+                            void *lpEnvironment,
+                            char *working_directory_utf8,
+                            LPSTARTUPINFOA si,
+                            LPPROCESS_INFORMATION lpProcessInformation)
+{
+  BOOL ret;
+  wchar_t *application_name = utf8_to_wchar0 (application_name_utf8);
+  wchar_t *command_line = utf8_to_wchar0 (command_line_utf8);
+  wchar_t *working_directory = utf8_to_wchar0 (working_directory_utf8);
+
+  STARTUPINFOW siw;
+  memset (&siw, 0, sizeof siw);
+  if (si)
+    {
+      siw.cb = sizeof (siw);
+      siw.dwFlags = si->dwFlags;
+      siw.wShowWindow = si->wShowWindow;
+      siw.hStdInput = si->hStdInput;
+      siw.hStdOutput = si->hStdOutput;
+      siw.hStdError = si->hStdError;
+      siw.dwX = si->dwX;
+      siw.dwY = si->dwY;
+      siw.dwXSize = si->dwXSize;
+      siw.dwYSize = si->dwYSize;
+      siw.dwXCountChars = si->dwXCountChars;
+      siw.dwYCountChars = si->dwYCountChars;
+      siw.dwFillAttribute = si->dwFillAttribute;
+      siw.lpDesktop = utf8_to_wchar0 (si->lpDesktop);
+      siw.lpTitle = utf8_to_wchar0 (si->lpTitle);
+    }
+
+  ret = CreateProcessW (application_name,
+                        command_line,
+                        lpProcessAttributes,
+                        lpThreadAttributes,
+                        bInheritHandles,
+                        dwCreationFlags,
+                        lpEnvironment,
+                        working_directory,
+                        si ? &siw : NULL,
+                        lpProcessInformation);
+  free (siw.lpTitle);
+  free (siw.lpDesktop);
+  free (application_name);
+  free (command_line);
+  free (working_directory);
+  return ret;
+}
 
 /* Entry point called by the DLL loader.  */
 #ifdef DLL_EXPORT
